package lightcontainer.domains.client;

import lightcontainer.domains.Task;
import lightcontainer.interfaces.ProtocolRepository;
import lightcontainer.interfaces.StoreProcessorFFE;
import lightcontainer.protocol.ProtocolReader;
import lightcontainer.protocol.ProtocolWriter;

import java.io.*;
import java.net.Socket;
import java.net.SocketException;
import java.nio.charset.StandardCharsets;
import java.util.Objects;

/**
 * StoreProcessor
 *
 * <!> MULTICAST CLIENT <!>
 * Class communicating with the storebackend and
 * processing the tasks in the FileFrontEnd
 *
 * @version     1.1
 * @since       1.0
 *
 * @see         Runnable
 * @see         AutoCloseable
 * @author      Jérémi NIHART <j.nihart@student.helmo.be>
 */
public class StoreProcessor extends Thread implements AutoCloseable {
    // Variables
    private final StoreProcessorFFE fileFrontEnd;
    private final Socket store;
    private final String domain;
    private boolean client_run;

    private BufferedReader reader;
    private Context context;
    private PrintWriter writer;
    private ProtocolWriter.ProtocolResult protocolResult;
    private ProtocolRepository protocolRep;

    // Constructor
    public StoreProcessor(Socket socket, String domain, StoreProcessorFFE ffe, ProtocolRepository protocolRep) {
        this.domain = domain;
        this.fileFrontEnd = ffe;
        this.store = socket;
        this.protocolRep = protocolRep;
        this.client_run = false;
    }

    /**
     * Initialise the Store's Reader and Writer.
     *
     * @see BufferedReader
     * @see PrintWriter
     * @since 1.0
     */
    public void startStore() {
        try {
            this.reader = new BufferedReader(new InputStreamReader(
                    this.store.getInputStream(),
                    StandardCharsets.UTF_8
            ));
            this.writer = new PrintWriter(new OutputStreamWriter(
                    this.store.getOutputStream(),
                    StandardCharsets.UTF_8
            ), true);

            this.start();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    /**
     * Thread Function
     * Start the dialogue with the storebackend.
     *
     * @since 1.0
     */
    @Override
    public void run() {
        this.client_run = true;
        while (this.client_run) {
            try {
                waitAction();
                System.out.println("[SBE] Envoie commande : " + protocolResult.getCommand());

                // Request
                this.writer.write(protocolResult.getCommand());
                this.writer.flush();

                try {
                    protocolResult.write(this.store.getOutputStream());
                } catch (IOException writeException) { // Si SBE fermé
                    System.out.println("STOPPER");
                    // Envoie au client que la requête n'a pu être traitée
                    alertAvalaible(null);
                    break;
                }

                // Response
                String responseCommand = this.reader.readLine();
                if (responseCommand != null && !responseCommand.isBlank()) {
                    responseCommand += "\r\n";
                    ProtocolReader.ProtocolResult responseResult = protocolRep.executeReader(context, responseCommand);
                    if (responseResult != null) {
                        System.out.println("StoreBackEnd response to client: " + responseResult.getResultCommand());
                        responseResult.read(
                                this.store.getInputStream()
                        );

                        alertAvalaible(responseResult.getResultCommand());
                    } else {
                        System.out.println("StoreBackEnd result: Commande null");
                        alertAvalaible(null);
                    }
                } else {
                    System.out.println("StoreBackEnd: Commande null");
                    alertAvalaible(null);
                }

            } catch (IOException exception) {
                System.out.println("[ERROR] Problem with SBE (" + domain + ") : " + exception.getMessage());
                this.close();
            }
        }

        // Fermeture du SBE
        try {
            this.reader.close();
            this.writer.close();
            this.store.close();
            this.fileFrontEnd.onStoreDisconnect(this.domain);
        } catch (IOException ioException) {
            System.out.println("[ERROR] Error while closing SBE (" + domain + ") : " + ioException.getMessage());
        }
    }

    /**
     * Permet de demander au StoreBackEnd d'effectuer une commande
     * @param protocolResult La commande à effectuer
     */
    public void executeCommand(Context context, ProtocolWriter.ProtocolResult protocolResult) {
        synchronized (this) {
            this.protocolResult = protocolResult;
            this.context = context;
            this.notify();
        }
    }

    /**
     * Permet de déclarer le StoreBackEnd disponible
     */
    private void alertAvalaible(ProtocolWriter.ProtocolResult responseCommand) {
        synchronized (this) {
            this.protocolResult = null;
            if (responseCommand == null) {
                this.close();
            }
            fileFrontEnd.onStoreAvailable(this, responseCommand);
        }
    }

    /**
     * Permet au Store d'attendre une action à réaliser
     */
    private void waitAction() {
        synchronized (this) {
            try {
                this.wait();
            } catch (InterruptedException e) { e.printStackTrace(); }
        }
    }

    /**
     * AutoClosable Function
     * Close the Storage thread and resources.
     *
     * @since 1.0
     */
    @Override
    public void close() {
        if (this.client_run) {
            this.client_run = false;
            // TODO : Gérer déconnection (enlever du repo et prévenir client et FileFrontEnd)
        }
    }

    /**
     * Compare two StoreProcessor object to check if equals.
     */
    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;
        StoreProcessor that = (StoreProcessor) o;
        return Objects.equals(domain, that.domain);
    }

    /**
     * Get hash code
     */
    @Override
    public int hashCode() {
        return Objects.hash(domain);
    }

    public boolean canProcessTask(Task task) {
        return this.protocolResult == null; // Vérifier si tâche veut ce SBE
    }

    public String getDomain() {
        return this.domain;
    }
}
